package eu.dnetlib.wfs.nodes;

import java.time.LocalDateTime;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Stream;

import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.DocumentHelper;
import org.springframework.beans.factory.annotation.Autowired;

import eu.dnetlib.common.clients.DnetServiceClientFactory;
import eu.dnetlib.common.clients.DsmClient;
import eu.dnetlib.common.clients.MDStoreManagerClient;
import eu.dnetlib.common.mdstores.backends.sql.MDStoreSqlBackend;
import eu.dnetlib.domain.dsm.Api;
import eu.dnetlib.domain.dsm.Datasource;
import eu.dnetlib.domain.mdstore.MDStoreVersion;
import eu.dnetlib.domain.mdstore.MDStoreWithInfo;
import eu.dnetlib.domain.mdstore.records.MetadataRecord;
import eu.dnetlib.wfs.annotations.WfInputParam;
import eu.dnetlib.wfs.annotations.WfNode;
import eu.dnetlib.wfs.collector.CollectorService;
import eu.dnetlib.wfs.utils.MdBuilderFactory;
import eu.dnetlib.wfs.utils.XpathFilterFactory;
import jakarta.transaction.Transactional;

@WfNode("md_collect_incremental")
public class MdCollectIncrementalJobNode extends AbstractJobNode {

	@WfInputParam
	private Datasource ds;

	@WfInputParam
	private Api api;

	@WfInputParam
	private String mdId;

	@WfInputParam
	private LocalDateTime overrideFromDate;

	@WfInputParam
	private LocalDateTime overrideUntilDate;

	@WfInputParam
	private String filterXpath;

	@Autowired
	private DnetServiceClientFactory clientFactory;

	@Autowired
	private CollectorService collectorService;

	@Autowired
	private MDStoreSqlBackend mdStoreSqlBackend;

	@Override
	@Transactional
	protected void execute() throws Exception {

		final LocalDateTime fromDate = overrideFromDate != null ? overrideFromDate : findLastCollDate(api);
		final LocalDateTime untilDate = overrideUntilDate != null ? overrideUntilDate : null;

		final Predicate<Document> filter = XpathFilterFactory.createFilter(filterXpath);
		final Function<Document, MetadataRecord> mdBuilder = MdBuilderFactory.createMdBuilder(ds, api);

		final MDStoreManagerClient mdstoreManager = clientFactory.getClient(MDStoreManagerClient.class);

		final MDStoreWithInfo outputMDStore = mdstoreManager.findMDStore(mdId);
		final MDStoreVersion outputVersion = mdstoreManager.findVersion(outputMDStore.getCurrentVersion());
		try {
			final Stream<MetadataRecord> stream = collectorService.collect(api, fromDate, untilDate)
					.map(xml -> {
						try {
							return DocumentHelper.parseText(xml);
						} catch (final DocumentException e) {
							throw new RuntimeException("Invalid record: " + xml);
						}
					})
					.filter(filter)
					.map(mdBuilder);

			mdStoreSqlBackend.saveRecords(outputVersion, stream);

			final long size = mdStoreSqlBackend.countRecords(outputVersion.getId());
			outputVersion.setSize(size);
			mdstoreManager.commitVersion(outputVersion);

			clientFactory.getClient(DsmClient.class).updateApiCollectionInfo(api.getId(), mdId, size);
		} catch (final Throwable e) {
			mdstoreManager.abortVersion(outputVersion);
			throw e;
		}
	}

	private LocalDateTime findLastCollDate(final Api api) {
		return api.getLastCollectionDate() != null ? api.getLastCollectionDate().toLocalDateTime() : null;
	}

}
