package eu.dnetlib.wfs.manager.controller;

import java.time.LocalDate;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;

import eu.dnetlib.common.controller.DnetRestController;
import eu.dnetlib.domain.common.KeyValue;
import eu.dnetlib.domain.wfs.WfConfiguration;
import eu.dnetlib.domain.wfs.WfJournalEntry;
import eu.dnetlib.domain.wfs.WfRepoHiDesc;
import eu.dnetlib.domain.wfs.WfRepoHiParams;
import eu.dnetlib.domain.wfs.WfSection;
import eu.dnetlib.domain.wfs.WfSubscription;
import eu.dnetlib.domain.wfs.WfTemplate;
import eu.dnetlib.domain.wfs.WorkflowsConstants;
import eu.dnetlib.wfs.manager.service.WorkflowManagerService;

@RestController
@RequestMapping("/api")
public class ApiController extends DnetRestController {

	@Autowired
	private WorkflowManagerService wfManagerService;

	@GetMapping("/history")
	public List<WfJournalEntry> history(
			@RequestParam(required = true) final int total,
			@RequestParam(required = false) final LocalDate from,
			@RequestParam(required = false) final LocalDate to) {
		return wfManagerService.findHistory(total, from, to);
	}

	@GetMapping("/history/byDsId/{dsId}")
	public List<WfJournalEntry> historyByDsId(@PathVariable final String dsId) {
		return wfManagerService.recentHistoryForDsId(dsId);
	}

	@GetMapping("/history/byApiId/{apiId}")
	public List<WfJournalEntry> historyByApiId(@PathVariable final String apiId) {
		return wfManagerService.recentHistoryForApiId(apiId);
	}

	@GetMapping("/history/byConf/{wfConfId}")
	public List<WfJournalEntry> history(@PathVariable final String wfConfId) {
		return wfManagerService.recentHistoryForConfiguration(wfConfId);
	}

	@GetMapping("/proc/{processId}")
	public List<WfJournalEntry> getProcessExecution(@PathVariable final String processId) {
		return Arrays.asList(wfManagerService.findProcessLog(processId));
	}

	@GetMapping("/sections")
	public Iterable<WfSection> listWfSections() throws Exception {
		return wfManagerService.listSections();
	}

	@GetMapping("/sections/{section}")
	public List<KeyValue<String>> listWfConfigurations(@PathVariable final String section) throws Exception {
		return wfManagerService.listWfConfigurationsBySection(section)
				.stream()
				.map(x -> new KeyValue<>(x.getId(), x.getName()))
				.collect(Collectors.toList());
	}

	@GetMapping("/conf/{id}")
	public WfConfiguration getWfConfiguration(@PathVariable final String id) throws Exception {
		return wfManagerService.findWorkflowConfiguration(id);
	}

	@PostMapping("/conf")
	public WfConfiguration saveWfConfiguration(@RequestBody final WfConfiguration conf) throws Exception {
		return wfManagerService.saveWfConfiguration(conf);
	}

	@DeleteMapping("/conf/{id}")
	public void deleteWfConfiguration(@PathVariable final String id) throws Exception {
		wfManagerService.deleteWfConfiguration(id);
	}

	@GetMapping("/conf/{id}/start")
	public List<WfJournalEntry> startWorkflowConfiguration(@PathVariable final String id) throws Exception {
		final WfConfiguration conf = wfManagerService.findWorkflowConfiguration(id);

		wfManagerService.prepareNewJob(conf, false);

		if (StringUtils.isNotBlank(conf.getApiId())) { return wfManagerService.recentHistoryForApiId(conf.getApiId()); }
		if (StringUtils.isNotBlank(conf.getDsId())) { return wfManagerService.recentHistoryForDsId(conf.getDsId()); }
		if (StringUtils.isNotBlank(conf.getId())) { return wfManagerService.recentHistoryForConfiguration(conf.getId()); }

		return wfManagerService.recentHistory();
	}

	@GetMapping("/conf/{id}/destroy")
	public List<WfJournalEntry> destroyWorkflowConfiguration(@PathVariable final String id) throws Exception {
		final WfConfiguration conf = wfManagerService.findWorkflowConfiguration(id);

		wfManagerService.prepareNewJob(conf, true);

		if (StringUtils.isNotBlank(conf.getApiId())) { return wfManagerService.recentHistoryForApiId(conf.getApiId()); }
		if (StringUtils.isNotBlank(conf.getDsId())) { return wfManagerService.recentHistoryForDsId(conf.getDsId()); }
		if (StringUtils.isNotBlank(conf.getId())) { return wfManagerService.recentHistoryForConfiguration(conf.getId()); }

		return wfManagerService.recentHistory();

	}

	@GetMapping("/conf/{id}/subscriptions")
	public List<WfSubscription> listWorkflowSubscriptions(@PathVariable final String id) throws Exception {
		return wfManagerService.listSubscriptions(id);
	}

	@PostMapping("/conf/{id}/subscriptions")
	public void saveWorkflowSubscriptions(@PathVariable final String id, @RequestBody final List<WfSubscription> subscriptions) throws Exception {
		wfManagerService.saveSubscriptions(id, subscriptions);
	}

	@GetMapping("/repo-his")
	public List<WfRepoHiDesc> listRepoHIs() {
		return wfManagerService.listRepoHis();
	}

	@PostMapping("/repo-hi/{id}/start")
	public WfJournalEntry startRepoHi(@PathVariable final String id, @RequestBody final WfRepoHiParams params) {
		return wfManagerService
				.prepareNewJob(id, null, WorkflowsConstants.REPO_HI_FAMILY, WorkflowsConstants.REPO_HI_FAMILY, params.getDsId(), params.getDsName(), params
						.getApiId());
	}

	@GetMapping("/template/{id}")
	public WfTemplate findWfTemplate(@PathVariable final String id) {
		return wfManagerService.findWfTemplate(id);
	}

	@GetMapping("/confs/byApi/{id}")
	public List<WfConfiguration> listWfConfigurationsByApi(@PathVariable final String id) {
		return wfManagerService.listWfConfigurationsByApiId(id);
	}

}
