package eu.dnetlib.domain.wfs;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

import org.apache.commons.lang3.StringUtils;
import org.springframework.core.env.Environment;

import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.annotation.JsonInclude.Include;

@JsonInclude(Include.NON_NULL)
public class WfTemplate implements Serializable {

	private static final long serialVersionUID = 5919290887480115842L;

	private List<WfParam> parameters = new ArrayList<>();
	private List<Node> graph = new ArrayList<>();
	private Map<String, String> info = new LinkedHashMap<>();

	public List<Node> getGraph() {
		return graph;
	}

	public void setGraph(final List<Node> graph) {
		this.graph = graph;
	}

	public List<WfParam> getParameters() {
		return parameters;
	}

	public void setParameters(final List<WfParam> parameters) {
		this.parameters = parameters;
	}

	public Map<String, String> getInfo() {
		return info;
	}

	public void setInfo(final Map<String, String> info) {
		this.info = info;
	}

	public static enum WfParamType {
		STRING,
		NUMBER,
		DATE,
		BOOLEAN
	}

	@JsonInclude(Include.NON_NULL)
	public static class WfParam implements Serializable {

		private static final long serialVersionUID = 5885589803738655166L;

		private String name;
		private String description;
		private WfParamType type = WfParamType.STRING;
		private String defaultValue;
		private boolean required = true;

		public WfParam() {}

		public WfParam(final String name) {
			this.name = name;
			description = name;
		}

		public WfParam(final String name, final WfParamType type) {
			this.name = name;
			description = name;
			this.type = type;
		}

		public WfParam(final String name, final String description, final WfParamType type, final String defaultValue, final boolean required) {
			this.name = name;
			this.description = description;
			this.type = type;
			this.defaultValue = defaultValue;
			this.required = required;
		}

		public String getName() {
			return name;
		}

		public void setName(final String name) {
			this.name = name;
		}

		public String getDescription() {
			return description;
		}

		public void setDescription(final String description) {
			this.description = description;
		}

		public WfParamType getType() {
			return type;
		}

		public void setType(final WfParamType type) {
			this.type = type;
		}

		public String getDefaultValue() {
			return defaultValue;
		}

		public void setDefaultValue(final String defaultValue) {
			this.defaultValue = defaultValue;
		}

		public boolean isRequired() {
			return required;
		}

		public void setRequired(final boolean required) {
			this.required = required;
		}

	}

	@JsonInclude(Include.NON_NULL)
	public static class Node implements Serializable {

		private static final long serialVersionUID = -3695762832959801906L;

		private static final String regExRef = "\\$\\{(\\w*)\\}";

		private String name;
		private String type;
		private boolean start = false;
		private boolean join = false;
		private List<Arc> arcs = new ArrayList<>();
		private List<NodeInputParam> input = new ArrayList<>();
		private List<NodeOutputParam> output = new ArrayList<>();

		public Node() {}

		public Node(final String name, final boolean start) {
			this.name = name;
			this.start = start;
		}

		public Node(final String name) {
			this.name = name;
		}

		public Node(final String name, final String type) {
			this.name = name;
			this.type = type;
		}

		public String getName() {
			return name;
		}

		public void setName(final String name) {
			this.name = name;
		}

		public String getType() {
			return type;
		}

		public void setType(final String type) {
			this.type = type;
		}

		public boolean isStart() {
			return start;
		}

		public void setStart(final boolean start) {
			this.start = start;
		}

		public boolean isJoin() {
			return join;
		}

		public void setJoin(final boolean join) {
			this.join = join;
		}

		public List<Arc> getArcs() {
			return arcs;
		}

		public void setArcs(final List<Arc> arcs) {
			this.arcs = arcs;
		}

		public List<NodeInputParam> getInput() {
			return input;
		}

		public void setInput(final List<NodeInputParam> input) {
			this.input = input;
		}

		public List<NodeOutputParam> getOutput() {
			return output;
		}

		public void setOutput(final List<NodeOutputParam> output) {
			this.output = output;
		}

		public Map<String, Object> findEnvParams() {
			return input.stream()
					.filter(p -> StringUtils.isNotBlank(p.getEnv()))
					.collect(Collectors.toMap(NodeInputParam::getName, NodeInputParam::getEnv));
		}

		public Map<String, Object> calculateInitialParams(final Map<String, Object> globalParams, final Environment environment) {

			final Map<String, Object> map = new HashMap<>();
			for (final NodeInputParam p : input) {
				if (StringUtils.isBlank(p.getEnv())) {
					map.put(p.getName(), calculateSimpleValue(p, globalParams, environment));
				}
			}

			return map;
		}

		private Object calculateSimpleValue(final NodeInputParam p, final Map<String, Object> globalParams, final Environment environment) {
			String value = p.getValue();
			final String ref = p.getRef();
			final String prop = p.getProperty();

			if (StringUtils.isNotBlank(ref) && (globalParams.get(ref) != null)) { return globalParams.get(ref); }

			if (StringUtils.isNotBlank(value)) {
				final Matcher matcher = Pattern.compile(regExRef, Pattern.MULTILINE).matcher(value);
				while (matcher.find()) {
					final String rName = matcher.group(1);
					final Object rValue = globalParams.get(rName);

					value = value.replaceAll(Pattern.quote(matcher.group(0)), rValue.toString());
				}
				return value;
			}
			if (StringUtils.isNotBlank(prop)) { return environment.getProperty(prop); }

			return null;
		}

		public Map<String, String> outputToEnvMap() {
			final Map<String, String> map = new HashMap<>();
			for (final NodeOutputParam p : output) {
				if (StringUtils.isNotBlank(p.getEnv())) {
					map.put(p.getName(), p.getEnv());
				}
			}
			return map;
		}
	}

	@JsonInclude(Include.NON_NULL)
	public static class Arc implements Serializable {

		private static final long serialVersionUID = 7866138976929522262L;

		private String to;
		private String condition;

		public Arc() {}

		public Arc(final String to) {
			this.to = to;
		}

		public Arc(final String to, final String condition) {
			this.to = to;
			this.condition = condition;
		}

		public String getTo() {
			return to;
		}

		public void setTo(final String to) {
			this.to = to;
		}

		public String getCondition() {
			return condition;
		}

		public void setCondition(final String condition) {
			this.condition = condition;
		}
	}

	@JsonInclude(Include.NON_NULL)
	public static class NodeInputParam implements Serializable {

		private static final long serialVersionUID = 7815785723401725707L;

		private String name;

		/**
		 * a constant value
		 */
		private String value;

		/**
		 * a global param
		 */
		private String ref;

		/**
		 * a system property
		 */
		private String property;

		/**
		 * a property from the WorkFlow ENV
		 */
		private String env;

		public static NodeInputParam newRefParam(final String name, final String ref) {
			final NodeInputParam np = new NodeInputParam();
			np.setName(name);
			np.setRef(ref);
			return np;
		}

		public static NodeInputParam newSimpleParam(final String name, final String value) {
			final NodeInputParam np = new NodeInputParam();
			np.setName(name);
			np.setValue(value);
			return np;
		}

		public static NodeInputParam newPropertyParam(final String name, final String property) {
			final NodeInputParam np = new NodeInputParam();
			np.setName(name);
			np.setProperty(property);
			return np;
		}

		public static NodeInputParam newEnvParam(final String name, final String envProp) {
			final NodeInputParam np = new NodeInputParam();
			np.setName(name);
			np.setEnv(envProp);
			return np;
		}

		public String getName() {
			return name;
		}

		public void setName(final String name) {
			this.name = name;
		}

		public String getValue() {
			return value;
		}

		public void setValue(final String value) {
			this.value = value;
		}

		public String getRef() {
			return ref;
		}

		public void setRef(final String ref) {
			this.ref = ref;
		}

		public String getProperty() {
			return property;
		}

		public void setProperty(final String property) {
			this.property = property;
		}

		public String getEnv() {
			return env;
		}

		public void setEnv(final String env) {
			this.env = env;
		}

	}

	@JsonInclude(Include.NON_NULL)
	public static class NodeOutputParam implements Serializable {

		private static final long serialVersionUID = -7995475104154153199L;

		private String name;

		private String env;

		public String getName() {
			return name;
		}

		public void setName(final String name) {
			this.name = name;
		}

		public String getEnv() {
			return env;
		}

		public void setEnv(final String env) {
			this.env = env;
		}

	}
}
