package eu.dnetlib.wfs.nodes.launch;

import java.time.LocalDateTime;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.UUID;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import eu.dnetlib.common.app.ServiceStatusRegistry;
import eu.dnetlib.domain.wfs.JobStatus;
import eu.dnetlib.domain.wfs.WfConfiguration;
import eu.dnetlib.domain.wfs.WfJournalEntry;
import eu.dnetlib.wfs.annotations.WfInputParam;
import eu.dnetlib.wfs.annotations.WfNode;
import eu.dnetlib.wfs.nodes.ProcessNode;
import eu.dnetlib.wfs.procs.ProcessEngine;
import eu.dnetlib.wfs.procs.Token;
import eu.dnetlib.wfs.procs.WorkflowProcess;
import eu.dnetlib.wfs.utils.NodeCallback;
import eu.dnetlib.wfs.utils.ProcessCallback;
import eu.dnetlib.wfs.utils.WfProcessUtils;

@WfNode("launchWorkflow")
public class LaunchWorkflowJobNode extends ProcessNode {

	private static final Log log = LogFactory.getLog(LaunchWorkflowJobNode.class);

	@WfInputParam
	private String wfId;

	private WorkflowProcess process;

	@Override
	public final void execute(final Token token, final NodeCallback nodeCallback) {
		try {
			final ProcessEngine engine = getEngine();

			final String procId = WfProcessUtils.generateProcessId();

			final WfConfiguration conf = new WfConfiguration();
			conf.setId("CHILD_" + UUID.randomUUID());
			conf.setParentId(process.getWfConfId());
			conf.setDetails(new HashMap<>());
			conf.setPriority(100);
			conf.setDsId(process.getDsId());
			conf.setDsName(process.getDsName());
			conf.setApiId(process.getDsInterface());
			conf.setEnabled(true);
			conf.setConfigured(true);
			conf.setSchedulingEnabled(false);
			conf.setCronExpression("");
			conf.setCronMinInterval(0);
			conf.setWorkflow(wfId);
			conf.setDestroyWf(null);
			conf.setSystemParams(process.getGlobalParams());
			conf.setUserParams(new HashMap<>());

			final WfJournalEntry job = new WfJournalEntry();
			job.setProcessId(procId);
			job.setWfConfigurationId(null);
			job.setName(process.getName() + ":" + getNodeName());
			job.setFamily(process.getFamily());

			job.setWfTemplateId(wfId);

			job.setDsId(process.getDsId());
			job.setDsName(process.getDsName());
			job.setDsApi(process.getDsInterface());

			job.setStatus(JobStatus.created);

			job.setDetails(new LinkedHashMap<>());
			job.setStartDate(null);
			job.setEndDate(null);
			job.setWfExecutor(ServiceStatusRegistry.getStatus().getName());
			job.setLastUpdate(LocalDateTime.now());

			engine.startWorkflowJob(job, conf, new ProcessCallback() {

				@Override
				public void onSuccess(final WorkflowProcess t) {
					log.debug("Child workflow has been completed successfully");
					nodeCallback.onSuccess(token);
				}

				@Override
				public void onFail(final WorkflowProcess t, final Throwable e) {
					log.error("Child workflow is failed");
					nodeCallback.onFail(token, e);
				}

			});

			if (log.isDebugEnabled()) {
				log.debug("The child workflow [conf: " + conf.getId() + "] is starting with procId: " + procId);
			}

			token.setProgressMessage("Launched sub workflow, proc: " + procId);

		} catch (final Throwable e) {
			log.error("got exception while launching child workflow", e);
			nodeCallback.onFail(token, e);
		}
	}

	@Override
	public void setProcess(final WorkflowProcess process) {
		this.process = process;
	}

	public String getWfId() {
		return wfId;
	}

	public void setWfId(final String wfId) {
		this.wfId = wfId;
	}

}
