package eu.dnetlib.wfs.procs;

import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.LinkedBlockingQueue;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.exception.ExceptionUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.postgresql.PGConnection;
import org.postgresql.PGNotification;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.ApplicationContext;
import org.springframework.jdbc.core.ConnectionCallback;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.stereotype.Service;

import eu.dnetlib.common.app.ServiceStatusRegistry;
import eu.dnetlib.common.clients.DnetServiceClientFactory;
import eu.dnetlib.common.clients.SimpleResourceClient;
import eu.dnetlib.domain.resource.SimpleResource;
import eu.dnetlib.domain.wfs.JobStatus;
import eu.dnetlib.domain.wfs.WfConfiguration;
import eu.dnetlib.domain.wfs.WfJournalEntry;
import eu.dnetlib.domain.wfs.WfTemplate;
import eu.dnetlib.domain.wfs.WorkflowsConstants;
import eu.dnetlib.errors.WorkflowManagerException;
import eu.dnetlib.wfs.graph.Graph;
import eu.dnetlib.wfs.graph.GraphLoader;
import eu.dnetlib.wfs.graph.GraphNode;
import eu.dnetlib.wfs.nodes.DefaultJobNode;
import eu.dnetlib.wfs.nodes.ProcessNode;
import eu.dnetlib.wfs.nodes.SuccessNode;
import eu.dnetlib.wfs.repository.WfConfigurationRepository;
import eu.dnetlib.wfs.repository.WfJournalEntryRepository;
import eu.dnetlib.wfs.utils.EmailSender;
import eu.dnetlib.wfs.utils.ProcessCallback;
import eu.dnetlib.wfs.utils.WfConfigurationUtils;
import jakarta.annotation.PostConstruct;
import jakarta.transaction.Transactional;

@Service
public class ProcessEngine {

	private static final Log log = LogFactory.getLog(ProcessEngine.class);

	@Autowired
	private EmailSender emailSender;

	@Autowired
	private WfJournalEntryRepository wfJournalEntryRepository;

	@Autowired
	private WfConfigurationRepository wfConfigurationRepository;

	@Autowired
	private JdbcTemplate jdbcTemplate;

	@Autowired
	private ApplicationContext applicationContext;

	@Autowired
	private ProcessRegistry processRegistry;

	@Autowired
	private DnetServiceClientFactory clientFactory;

	@Autowired
	private GraphLoader graphLoader;

	@Value("${dnet.wf.procs.size:20}")
	private int maxSize;

	private final LinkedBlockingQueue<String> queue = new LinkedBlockingQueue<>();

	@PostConstruct
	public void createNotificationHandler() {

		final Thread t1 = new Thread(this::notificationHandler, "wf-created-listener");
		t1.start();

		final Thread t2 = new Thread(this::consumeQueue, "wf-queue-consumer");
		t2.start();
	}

	private void notificationHandler() {
		while ((ServiceStatusRegistry.getStatus() == null) || StringUtils.isBlank(ServiceStatusRegistry.getStatus().getName())) {
			try {
				Thread.sleep(1000);
			} catch (final InterruptedException e) {
				throw new RuntimeException(e);
			}
		}

		final ConnectionCallback<Integer> action = c -> {
			c.createStatement().execute("LISTEN " + WorkflowsConstants.WF_CREATED_NOTIFICATION_CHANNEL);
			final PGConnection pgconn = c.unwrap(PGConnection.class);
			while (!Thread.currentThread().isInterrupted()) {
				final PGNotification[] nts = pgconn.getNotifications(10000);
				if (nts != null) {
					for (final PGNotification n : nts) {
						final String processId = n.getParameter();
						log.debug("NOTIFICATION RECEIVED: " + processId);
						queue.add(processId);
					}
				}
			}
			return 0;
		};

		jdbcTemplate.execute(action);
	}

	private void consumeQueue() {
		while ((ServiceStatusRegistry.getStatus() == null) || StringUtils.isBlank(ServiceStatusRegistry.getStatus().getName())) {
			try {
				Thread.sleep(1000);
			} catch (final InterruptedException e) {
				throw new RuntimeException(e);
			}
		}

		while (true) {
			try {
				tryToStartWf(queue.take(), ServiceStatusRegistry.getStatus().getName());
			} catch (final Throwable e) {
				log.error("Error accepting new wfs", e);
			}
		}
	}

	@Transactional
	private void tryToStartWf(final String processId, final String executor) {
		log.debug("Trying to start " + processId);

		wfJournalEntryRepository.tryAssegnment(processId, executor, JobStatus.accepted, LocalDateTime.now());

		final Optional<WfJournalEntry> job = wfJournalEntryRepository.findById(processId);

		if (job.isPresent() && (job.get().getStatus() == JobStatus.accepted) && job.get().getWfExecutor().equals(executor)) {
			startWorkflowJob(job.get(), null, null);
		}

	}

	@Transactional
	public void startWorkflowJob(final WfJournalEntry job, final WfConfiguration passedConf, final ProcessCallback callback) {

		try {

			log.debug("Configure process " + job.getProcessId() + " using wf template " + job.getWfTemplateId());

			final WfConfiguration conf;
			if (passedConf != null) {
				conf = passedConf;
			} else if (StringUtils.isNotBlank(job.getWfConfigurationId())) {
				conf = wfConfigurationRepository.findById(job.getWfConfigurationId())
						.orElseThrow(() -> new WorkflowManagerException("WF configuration not found: " + job.getWfConfigurationId()));
			} else {
				conf = new WfConfiguration();
				conf.setId(null);
				conf.setName(job.getFamily());
				conf.setDetails(new HashMap<>());
				conf.setPriority(100);
				conf.setDsId(job.getDsId());
				conf.setDsName(job.getDsName());
				conf.setApiId(job.getDsApi());
				conf.setEnabled(true);
				conf.setConfigured(true);
				conf.setSchedulingEnabled(false);
				conf.setCronExpression("");
				conf.setCronMinInterval(0);
				conf.setWorkflow(job.getWfTemplateId());
				conf.setDestroyWf(null);
				conf.setSystemParams(new HashMap<>());
				conf.setUserParams(new HashMap<>());
			}

			if (!conf.isEnabled() || !conf.isConfigured()) {
				log.warn("Wf configuration " + conf.getId() + " is not ready to start");
				throw new WorkflowManagerException("Wf configuration " + conf.getId() + " is not ready to start");
			}

			final SimpleResourceClient simpleResourceClient = clientFactory.getClient(SimpleResourceClient.class);

			final SimpleResource wfMetadata = simpleResourceClient.findResource(job.getWfTemplateId());
			if (!WorkflowsConstants.WF_TEMPLATE.equals(wfMetadata.getType())) { throw new WorkflowManagerException("WF not found: " + conf.getWorkflow()); }

			final WfTemplate wfTmpl = simpleResourceClient.findResourceContent(job.getWfTemplateId(), WfTemplate.class);

			final Map<String, Object> globalParams = WfConfigurationUtils.allConfiguredParameters(wfTmpl.getParameters(), conf);

			final Graph graph = graphLoader.loadGraph(wfTmpl, globalParams);

			final WorkflowProcess process = new WorkflowProcess(job.getProcessId(), wfMetadata, conf, graph, globalParams, callback);

			processRegistry.registerProcess(process, conf);

			job.setStatus(JobStatus.running);
			job.setLastUpdate(LocalDateTime.now());

			wfJournalEntryRepository.save(job);

			startProcess(process);

		} catch (final Throwable e) {
			final Map<String, String> details = new LinkedHashMap<>();
			details.put(WorkflowsConstants.LOG_SYSTEM_ERROR, e.getMessage());
			details.put(WorkflowsConstants.LOG_SYSTEM_ERROR_STACKTRACE, ExceptionUtils.getStackTrace(e));
			job.setDetails(details);
			job.setStatus(JobStatus.failure);
			job.setEndDate(LocalDateTime.now());
			job.setLastUpdate(LocalDateTime.now());

			wfJournalEntryRepository.save(job);
		}
	}

	public final void killProcess(final String procId) {
		processRegistry.findProcess(procId).kill();
	}

	public void startProcess(final WorkflowProcess process) {

		log.info("Starting process: " + process);
		log.debug(process.getGraph());

		final LocalDateTime now = LocalDateTime.now();
		process.setStatus(JobStatus.running);
		process.setStartDate(now);
		process.setLastActivityDate(now);

		try {
			for (final GraphNode graphNode : process.getGraph().startNodes()) {
				final Token token = process.newToken(process.getGlobalParams());
				executeNode(process, graphNode, token);
			}
		} catch (final Throwable e) {
			log.error("WorkflowProcess node instantiation failed", e);
			process.setStatus(JobStatus.failure);
		}
	}

	public void releaseToken(final WorkflowProcess process, final GraphNode oldGraphNode, final Token oldToken) {
		process.setLastActivityDate(LocalDateTime.now());

		try {
			for (final GraphNode graphNode : process.getGraph().nextNodes(oldGraphNode, oldToken)) {
				if (graphNode.isJoin() || graphNode.isSuccessNode()) {
					if (!process.getPausedJoinNodeTokens().containsKey(graphNode.getName())) {
						process.getPausedJoinNodeTokens().put(graphNode.getName(), new ArrayList<>());
					}

					final List<Token> list = process.getPausedJoinNodeTokens().get(graphNode.getName());

					list.add(oldToken);

					if (list.size() == process.getGraph().getNumberOfIncomingArcs(graphNode)) {
						final Token token = process.newToken(mergeEnvParams(list.toArray(new Token[list.size()])));

						if (graphNode.isSuccessNode()) {
							completeProcess(process, token);
						} else {
							executeNode(process, graphNode, token);
						}
					}
				} else {
					final Token token = process.newToken(oldToken.getEnv());
					executeNode(process, graphNode, token);
				}
			}
		} catch (

		final Throwable e) {
			log.error("WorkflowProcess node instantiation failed", e);
			process.setStatus(JobStatus.failure);
			process.setError(e.getMessage());
			process.setErrorStacktrace(ExceptionUtils.getStackTrace(e));
			process.setLastActivityDate(LocalDateTime.now());
		}

	}

	private void executeNode(final WorkflowProcess process, final GraphNode graphNode, final Token token) throws WorkflowManagerException {
		process.setLastActivityDate(LocalDateTime.now());

		final Map<String, Object> params = new HashMap<>();
		if (graphNode.getParams() != null) {
			graphNode.getParams().forEach((k, v) -> params.put(k, v));
		}
		if (graphNode.getEnvParams() != null) {
			graphNode.getEnvParams().forEach((k, v) -> params.put(k, token.getEnvAttribute(v.toString())));
		}

		final ProcessNode pNode = newProcessNode(graphNode, process);

		pNode.execute(token, params);

	}

	private ProcessNode newProcessNode(final GraphNode graphNode, final WorkflowProcess process) throws WorkflowManagerException {
		if (graphNode.isSuccessNode()) { return new SuccessNode(); }
		if (StringUtils.isBlank(graphNode.getType())) { return new DefaultJobNode(graphNode.getName()); }
		final ProcessNode pnode = (ProcessNode) applicationContext.getBean(graphNode.getType());
		if (pnode == null) {
			log.error("cannot find bean of type " + graphNode.getType());
			throw new WorkflowManagerException("cannot find bean of type " + graphNode.getType());
		}
		pnode.setNodeName(graphNode.getName());
		pnode.setProcess(process);
		pnode.setGraphNode(graphNode);
		pnode.setEngine(this);

		return pnode;

	}

	private Map<String, Object> mergeEnvParams(final Token... tokens) {
		final Map<String, Object> map = new HashMap<>();
		Arrays.stream(tokens).forEach(t -> map.putAll(t.getEnv()));
		return map;
	}

	public void completeProcess(final WorkflowProcess process, final Token token) {
		process.complete(token);

		log.debug("Process completed " + process.getId());

		final WfJournalEntry job = process.asLog();
		wfJournalEntryRepository.save(job);

		emailSender.sendMails(process);
	}

}
