package eu.dnetlib.wfs.procs;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.TimeUnit;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Service;

import eu.dnetlib.domain.wfs.WfConfiguration;
import eu.dnetlib.errors.WorkflowManagerException;

@Service
public class ProcessRegistry {

	private static final Log log = LogFactory.getLog(ProcessRegistry.class);
	private final Map<String, WorkflowProcess> procs = new HashMap<>();
	private final Map<String, WorkflowProcess> byConfId = new HashMap<>();

	synchronized public int countRunningWfs() {
		int count = 0;
		for (final Map.Entry<String, WorkflowProcess> e : procs.entrySet()) {
			final WorkflowProcess proc = e.getValue();
			if (!proc.isTerminated()) {
				count++;
			}
		}
		return count;
	}

	public WorkflowProcess findProcess(final String procId) {
		return procs.get(procId);
	}

	public Collection<WorkflowProcess> listProcesses() {
		return procs.values();
	}

	public WorkflowProcess findProcsByConfigurationId(final String id) {
		return byConfId.get(id);
	}

	public void registerProcess(final WorkflowProcess process, final WfConfiguration conf) throws WorkflowManagerException {
		if (procs.containsValue(process) || procs.containsKey(process.getId())) {
			log.error("Already registerd process: " + process);
			throw new WorkflowManagerException("Already registered process: " + process);
		}
		procs.put(process.getId(), process);
		byConfId.put(conf.getId(), process);
	}

	@Scheduled(fixedRate = 10, timeUnit = TimeUnit.MINUTES)
	public void removeTerminatedProcess() {
		for (final Map.Entry<String, WorkflowProcess> e : procs.entrySet()) {
			final WorkflowProcess proc = e.getValue();
			if (proc.isTerminated()) {
				unregisterProcess(proc.getId());
			}
		}
	}

	public void unregisterProcess(final String procId) {
		synchronized (this) {
			final WorkflowProcess process = procs.remove(procId);
			if (process != null) {
				byConfId.entrySet()
						.stream()
						.filter(e -> e.getValue().getId().equals(process.getId()))
						.map(Entry::getKey)
						.forEach(confId -> byConfId.remove(confId, process));
			}
		}
	}

}
