package eu.dnetlib.wfs.procs;

import java.time.LocalDateTime;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CopyOnWriteArrayList;

import org.apache.commons.lang3.exception.ExceptionUtils;
import org.apache.commons.lang3.math.NumberUtils;

import eu.dnetlib.domain.resource.SimpleResource;
import eu.dnetlib.domain.wfs.JobStatus;
import eu.dnetlib.domain.wfs.WfConfiguration;
import eu.dnetlib.domain.wfs.WfJournalEntry;
import eu.dnetlib.domain.wfs.WorkflowsConstants;
import eu.dnetlib.wfs.graph.Graph;
import eu.dnetlib.wfs.utils.ProcessCallback;

/**
 * Created by michele on 19/11/15.
 */
public class WorkflowProcess implements Comparable<WorkflowProcess> {

	public enum StartMode {
		AUTO,
		MANUAL,
		DISABLED
	}

	private final String id;
	private final SimpleResource wfMetadata;
	private final WfConfiguration wfConf;
	private final Graph graph;
	private final ProcessCallback callback;
	private final List<Token> tokens = new CopyOnWriteArrayList<>();
	private LocalDateTime lastActivityDate;
	private JobStatus status;
	private LocalDateTime startDate = LocalDateTime.MIN;
	private LocalDateTime endDate = LocalDateTime.MIN;
	private final Map<String, List<Token>> pausedJoinNodeTokens = new HashMap<>();
	private final Map<String, Object> globalParams;
	private String error;
	private String errorStacktrace;
	private final Map<String, String> outputParams = new HashMap<>();

	public WorkflowProcess(
			final String id,
			final SimpleResource wfMetadata,
			final WfConfiguration wfConf,
			final Graph graph,
			final Map<String, Object> globalParams,
			final ProcessCallback callback) {
		this.id = id;
		this.wfMetadata = wfMetadata;
		this.wfConf = wfConf;
		this.graph = graph;
		this.callback = callback;
		status = JobStatus.created;
		this.globalParams = globalParams;
		lastActivityDate = LocalDateTime.now();
	}

	public String getId() {
		return id;
	}

	public String getName() {
		return wfConf.getName();
	}

	public String getFamily() {
		return wfMetadata.getSubtype();
	}

	public String getWfId() {
		return wfMetadata.getId();
	}

	public String getWfConfId() {
		return wfConf.getId();
	}

	public String getParentId() {
		return wfConf.getParentId();
	}

	public int getPriority() {
		return wfConf.getPriority();
	}

	public String getDsId() {
		return wfConf.getId();
	}

	public String getDsName() {
		return wfConf.getDsName();
	}

	public String getDsInterface() {
		return wfConf.getApiId();
	}

	public Map<String, List<Token>> getPausedJoinNodeTokens() {
		return pausedJoinNodeTokens;
	}

	public JobStatus getStatus() {
		return status;
	}

	public void setStatus(final JobStatus status) {
		this.status = status;
	}

	public Graph getGraph() {
		return graph;
	}

	public List<Token> getTokens() {
		return tokens;
	}

	public void kill() {
		setStatus(JobStatus.killed);
	}

	public boolean isTerminated() {
		return switch (status) {
		case success, failure, killed -> true;
		default -> false;
		};
	}

	public LocalDateTime getLastActivityDate() {
		return lastActivityDate;
	}

	public void setLastActivityDate(final LocalDateTime lastActivityDate) {
		this.lastActivityDate = lastActivityDate;
	}

	@Override
	public String toString() {
		return String.format("[process id='%s' name='%s']", id, wfMetadata.getName());
	}

	@Override
	public int compareTo(final WorkflowProcess wp) {
		return NumberUtils.compare(getPriority(), wp.getPriority());
	}

	public Map<String, Object> getGlobalParams() {
		return globalParams;
	}

	public void setStartDate(final LocalDateTime startDate) {
		this.startDate = startDate;
	}

	public void setEndDate(final LocalDateTime endDate) {
		this.endDate = endDate;
	}

	public LocalDateTime getStartDate() {
		return startDate;
	}

	public LocalDateTime getEndDate() {
		return endDate;
	}

	public String getError() {
		return error;
	}

	public void setError(final String error) {
		this.error = error;
	}

	public String getErrorStacktrace() {
		return errorStacktrace;
	}

	public void setErrorStacktrace(final String errorStacktrace) {
		this.errorStacktrace = errorStacktrace;
	}

	public Map<String, String> getOutputParams() {
		return outputParams;
	}

	public Token newToken(final Map<String, Object> attrs) {
		final Token token = new Token();
		token.addEnvAttributes(attrs);
		tokens.add(token);
		return token;
	}

	public void complete(final Token token) {
		final LocalDateTime now = token.getEndDate();
		setLastActivityDate(now);
		setEndDate(now);
		setStatus(token.isFailed() ? JobStatus.failure : JobStatus.success);

		if (token.isFailed()) {
			setStatus(JobStatus.failure);
			setError(token.getError().getMessage());
			setErrorStacktrace(ExceptionUtils.getStackTrace(token.getError()));
			setLastActivityDate(LocalDateTime.now());
		}

		if (callback != null) {
			if (token.isFailed()) {
				callback.onFail(this, token.getError());
			} else {
				callback.onSuccess(this);
			}
		}

	}

	public WfJournalEntry asLog() {
		final Map<String, String> details = new LinkedHashMap<>();
		details.putAll(getOutputParams());
		details.put(WorkflowsConstants.LOG_WF_PRIORITY, "" + getPriority());
		details.put(WorkflowsConstants.LOG_WF_ID, getWfId());
		details.put(WorkflowsConstants.LOG_WF_CONF_ID, getWfConfId());

		if (getError() != null) {
			details.put(WorkflowsConstants.LOG_SYSTEM_ERROR, getError());
			details.put(WorkflowsConstants.LOG_SYSTEM_ERROR_STACKTRACE, getErrorStacktrace());
		}

		final WfJournalEntry pe = new WfJournalEntry();
		pe.setProcessId(getId());
		pe.setWfTemplateId(getWfId());
		pe.setWfConfigurationId(getWfConfId());
		pe.setName(getName());
		pe.setFamily(getFamily());

		pe.setDsId(getDsId());
		pe.setDsName(getDsName());
		pe.setDsApi(getDsInterface());

		pe.setStartDate(getStartDate());
		pe.setEndDate(getEndDate());
		pe.setStatus(getStatus());

		pe.setDetails(details);

		return pe;
	}

}
